"""
### Py Jaxtyping
> Pydantic support for Jaxtyping array annotations

#### Usage
  
Instead of:
>>> Int[np.ndarray, 'B 256 64 3']
do:
>>> PyArray[Int, int, 'B 256 64 3']

You can use it with `jaxtyping` as normal, but also it will:
- Serialize to nested lists
- Validate the correct shape and datatypes from serialized lists

### Example
```
from pydantic import BaseModel, ConfigDict
from py_jaxtyping import PyArray
from jaxtyping import Int
import numpy as np

class Sample(BaseModel):
  img: PyArray[Int, int, "W H 3"]
  label: str

Sample.model_validate({
  'img': np.ones((256, 64, 3), dtype=int),
  'label': 'car'
})
# checks out!


Sample.model_validate({
  'img': np.ones((256, 64, 1), dtype=int),
  'label': 'car'
})
# fails: invalid dims :/
```
"""
from .main import PyArray
from .schemas import array_schema, json_schema, core_schema


from pydantic_core.core_schema import no_info_after_validator_function, plain_serializer_function_ser_schema
import numpy as np
from .schemas import core_schema

def str_type(DType: type[int] | type[float] | type[bool]):
  if DType is int:
    return 'int'
  if DType is float:
    return 'float'
  if DType is int:
    return 'bool'

class PyArray(np.ndarray):
  """A `jaxtyping` array that can be used with pydantic
  
  Instead of:
  >>> Int[np.ndarray, 'B 256 64 3']
  do:
  >>> PyArray[Int, 'B 256 64 3']

  You can use it with `jaxtyping` as normal, but also

  ```
  class WithArray(BaseModel):
    arr: PyArray[Int, int, 'B 256 64 3']
  ```

  And it will:
  - Serialize to nested lists
  - Validate the correct shape and datatypes from serialized lists
  """
  def __class_getitem__(cls, params: tuple[type, type[int] | type[bool] | type[float], str]):
    DType, dtype, dims = params
    Type = DType[np.ndarray, dims]

    class Cls(Type):
      __name__ = f'Py{DType.__name__}Array[{dims}]'
      __qualname__ = __name__

      @classmethod
      def __get_pydantic_core_schema__(cls, *_):
        def validate(xs: list):
          arr = np.array(xs)
          if not isinstance(arr, Type):
            raise ValueError('Invalid array')
          return arr

        return no_info_after_validator_function(
          validate,
          core_schema(dims.split(' '), str_type(dtype)),
          serialization=plain_serializer_function_ser_schema(lambda x: x.tolist())
        ) # type: ignore

    return Cls


from typing import Annotated, Any, Unpack, Literal, TypedDict, TypeVar
from pydantic_core import CoreSchema

class SchemaKeys(TypedDict):
  array: str
  dtype: str
  items: str
  minItems: str
  maxItems: str

def array_schema(dims: list, **keys: Unpack[SchemaKeys]) -> dict[str, Any]:
  """Schema for a `len(dims)`-dimensional array
  - `dims[i]` can be
    - an integer: to denote a fixed shape
    - anything else: to denote a variable shape

  e.g. `dims=['N', 2, '5']`
  """
  match dims:
    case [dim] if str(dim).isdecimal():
      d = int(dim)
      return {
        'type': keys['array'], keys['items']: array_schema([], **keys),
        keys['minItems']: d, keys['maxItems']: d
      }
    case [dim]:
      return {
        'type': keys['array'], keys['items']: array_schema([], **keys),
      }
    case [dim, *ds] if str(dim).isdecimal():
      d = int(dim)
      return {
        'type': keys['array'], keys['items']: array_schema(ds, **keys),
        keys['minItems']: d, keys['maxItems']: d
      }
    case [dim, *ds]:
      return {
        'type': keys['array'], keys['items']: array_schema(ds, **keys),
      }
    case _:
      return { 'type': keys['dtype'] }


def json_schema(dims: list) -> dict[str, Any]:
  return array_schema(
    dims, array='array', dtype='number', items='items',
    minItems='minItems', maxItems='maxItems'
  )

def core_schema(dims: list, dtype: Literal['int', 'float', 'bool'] = 'int') -> CoreSchema:
  return array_schema(
    dims, array='list', dtype=dtype, items='items_schema',
    minItems='min_length', maxItems='max_length'
  ) # type: ignore